@@ -22,6 +22,8 @@ enum log_entry_type {
     LOGGER_SLAB_MOVE,
     LOGGER_CONNECTION_NEW,
     LOGGER_CONNECTION_CLOSE,
+    LOGGER_CONNECTION_ERROR,
+    LOGGER_CONNECTION_TLSERROR,
     LOGGER_DELETIONS,
 #ifdef EXTSTORE
     LOGGER_EXTSTORE_WRITE,
@@ -127,6 +129,7 @@ struct logentry_conn_event {
 struct logentry_proxy_req {
     unsigned short type;
     unsigned short code;
+    uint8_t flag;
     int status;
     int conn_fd;
     uint32_t reqlen;
