@@ -41,3 +41,8 @@ extern uint64_t ntohll(uint64_t);
  */
 void vperror(const char *fmt, ...)
     __gcc_attribute__ ((format (printf, 1, 2)));
+
+/* Some common timepsec functions.
+ */
+
+void mc_timespec_add(struct timespec *ts1, struct timespec *ts2);
